@@ -309,7 +309,6 @@ main (int    argc,
         GMainLoop          *main_loop;
         GOptionContext     *context;
         GError             *error = NULL;
-        int                 ret;
         gboolean            res;
         static gboolean     do_timed_exit    = FALSE;
         static gboolean     print_version    = FALSE;
@@ -328,8 +327,6 @@ main (int    argc,
         textdomain (GETTEXT_PACKAGE);
         setlocale (LC_ALL, "");
 
-        ret = 1;
-
         context = g_option_context_new (_("GNOME Display Manager"));
         g_option_context_add_main_entries (context, entries, NULL);
         g_option_context_set_ignore_unknown_options (context, TRUE);
@@ -338,14 +335,21 @@ main (int    argc,
         res = g_option_context_parse (context, &argc, &argv, &error);
         g_option_context_free (context);
         if (! res) {
-                g_warning ("%s", error->message);
+                g_printerr ("Failed to parse options: %s\n", error->message);
                 g_error_free (error);
-                goto out;
+                return EXIT_FAILURE;
         }
 
         if (print_version) {
                 g_print ("GDM %s\n", VERSION);
-                exit (1);
+                return EXIT_SUCCESS;
+        }
+
+        /* XDM compliant error message */
+        if (getuid () != 0) {
+                /* make sure the pid file doesn't get wiped */
+                g_printerr ("%s\n", _("Only the root user can run GDM"));
+                return EXIT_FAILURE;
         }
 
         if (fatal_warnings) {
@@ -359,14 +363,9 @@ main (int    argc,
         gdm_log_init ();
 
         settings = gdm_settings_new ();
-        if (settings == NULL) {
-                g_warning ("Unable to initialize settings");
-                goto out;
-        }
-
         if (! gdm_settings_direct_init (settings, DATADIR "/gdm/gdm.schemas", "/")) {
                 g_warning ("Unable to initialize settings");
-                goto out;
+                return EXIT_FAILURE;
         }
 
         gdm_log_set_debug (is_debug_set ());
@@ -375,13 +374,6 @@ main (int    argc,
 
         gdm_daemon_ensure_dirs (gdm_uid, gdm_gid);
 
-        /* XDM compliant error message */
-        if (getuid () != 0) {
-                /* make sure the pid file doesn't get wiped */
-                g_warning (_("Only the root user can run GDM"));
-                exit (-1);
-        }
-
         /* Connect to the bus, own the name and start the manager */
         bus_reconnect ();
 
@@ -413,14 +405,7 @@ main (int    argc,
 
         g_main_loop_unref (main_loop);
 
-        ret = 0;
-
- out:
-        if (error) {
-                g_printerr ("%s\n", error->message);
-                g_clear_error (&error);
-        }
-        return ret;
+        return EXIT_SUCCESS;
 }
 
 static void
@@ -434,7 +419,7 @@ on_name_acquired (GDBusConnection *bus,
         manager = gdm_manager_new ();
         if (manager == NULL) {
                 g_warning ("Could not construct manager object");
-                exit (1);
+                exit (EXIT_FAILURE);
         }
 
         g_debug ("Successfully connected to D-Bus");
