@@ -41,7 +41,10 @@
 struct GdmClientPrivate
 {
         GdmManager         *manager;
+
         GdmUserVerifier    *user_verifier;
+        GHashTable         *user_verifier_extensions;
+
         GdmGreeter         *greeter;
         GdmRemoteGreeter   *remote_greeter;
         GdmChooser         *chooser;
@@ -49,6 +52,7 @@ struct GdmClientPrivate
         char               *address;
 
         GList              *pending_opens;
+        char              **enabled_extensions;
 };
 
 static void     gdm_client_class_init  (GdmClientClass *klass);
@@ -104,6 +108,8 @@ on_got_manager (GdmManager          *manager,
         }
 
         g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
+        g_object_unref (client);
 }
 
 static void
@@ -126,6 +132,7 @@ get_manager (GdmClient           *client,
                                                            (GDestroyNotify)
                                                            g_object_unref);
                 g_simple_async_result_complete_in_idle (result);
+                g_object_unref (result);
                 return;
         }
 
@@ -139,18 +146,135 @@ get_manager (GdmClient           *client,
                                        result);
 }
 
+static void
+complete_user_verifier_proxy_operation (GdmClient          *client,
+                                        GSimpleAsyncResult *operation_result)
+{
+        g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
+}
+
+static void
+maybe_complete_user_verifier_proxy_operation (GdmClient          *client,
+                                              GSimpleAsyncResult *operation_result)
+{
+        GHashTableIter iter;
+        gpointer key, value;
+
+        if (client->priv->user_verifier_extensions != NULL) {
+                g_hash_table_iter_init (&iter, client->priv->user_verifier_extensions);
+                while (g_hash_table_iter_next (&iter, &key, &value)) {
+                        if (value == NULL)
+                                return;
+                }
+        }
+
+        complete_user_verifier_proxy_operation (client, operation_result);
+}
+
+static void
+on_user_verifier_choice_list_proxy_created (GObject            *source,
+                                            GAsyncResult       *result,
+                                            GSimpleAsyncResult *operation_result)
+{
+        GdmClient                 *client;
+        GdmUserVerifierChoiceList *choice_list;
+        GError                    *error = NULL;
+
+        client = GDM_CLIENT (g_async_result_get_source_object (G_ASYNC_RESULT (operation_result)));
+
+        choice_list = gdm_user_verifier_choice_list_proxy_new_finish (result, &error);
+
+        if (choice_list == NULL) {
+                g_debug ("Couldn't create UserVerifier ChoiceList proxy: %s", error->message);
+                g_clear_error (&error);
+                g_hash_table_remove (client->priv->user_verifier_extensions, gdm_user_verifier_choice_list_interface_info ()->name);
+        } else {
+                g_hash_table_replace (client->priv->user_verifier_extensions, gdm_user_verifier_choice_list_interface_info ()->name, choice_list);
+        }
+
+        maybe_complete_user_verifier_proxy_operation (client, operation_result);
+}
+
+static void
+on_user_verifier_extensions_enabled (GdmUserVerifier    *user_verifier,
+                                     GAsyncResult       *result,
+                                     GSimpleAsyncResult *operation_result)
+{
+        GdmClient *client;
+        GCancellable *cancellable;
+        GDBusConnection *connection;
+        GError    *error = NULL;
+        size_t     i;
+
+        client = GDM_CLIENT (g_async_result_get_source_object (G_ASYNC_RESULT (operation_result)));
+        cancellable = g_object_get_data (G_OBJECT (operation_result), "cancellable");
+
+        gdm_user_verifier_call_enable_extensions_finish (user_verifier, result, &error);
+
+        if (error != NULL) {
+                g_debug ("Couldn't enable user verifier extensions: %s",
+                         error->message);
+                g_clear_error (&error);
+                complete_user_verifier_proxy_operation (client, operation_result);
+                return;
+        }
+
+        connection = g_dbus_proxy_get_connection (G_DBUS_PROXY (user_verifier));
+
+        for (i = 0; client->priv->enabled_extensions[i] != NULL; i++) {
+                g_debug ("Enabled extensions[%lu] = %s", i, client->priv->enabled_extensions[i]);
+                g_hash_table_insert (client->priv->user_verifier_extensions, client->priv->enabled_extensions[i], NULL);
+
+                if (strcmp (client->priv->enabled_extensions[i],
+                            gdm_user_verifier_choice_list_interface_info ()->name) == 0) {
+                        g_hash_table_insert (client->priv->user_verifier_extensions, client->priv->enabled_extensions[i], NULL);
+                        gdm_user_verifier_choice_list_proxy_new (connection,
+                                                                 G_DBUS_PROXY_FLAGS_NONE,
+                                                                 NULL,
+                                                                 SESSION_DBUS_PATH,
+                                                                 cancellable,
+                                                                 (GAsyncReadyCallback)
+                                                                 on_user_verifier_choice_list_proxy_created,
+                                                                 operation_result);
+                } else {
+                        g_debug ("User verifier extension %s is unsupported", client->priv->enabled_extensions[i]);
+                        g_hash_table_remove (client->priv->user_verifier_extensions,
+                                             client->priv->enabled_extensions[i]);
+                }
+        }
+
+        if (g_hash_table_size (client->priv->user_verifier_extensions) == 0) {
+                g_debug ("No supported user verifier extensions");
+                complete_user_verifier_proxy_operation (client, operation_result);
+        }
+
+}
+
+static void
+free_interface_skeleton (GDBusInterfaceSkeleton *interface)
+{
+        if (interface == NULL)
+                return;
+
+        g_object_unref (interface);
+}
+
 static void
 on_user_verifier_proxy_created (GObject            *source,
                                 GAsyncResult       *result,
                                 GSimpleAsyncResult *operation_result)
 {
+        GdmClient       *self;
         GdmUserVerifier *user_verifier;
+        GCancellable    *cancellable = NULL;
         GError          *error = NULL;
 
         user_verifier = gdm_user_verifier_proxy_new_finish (result, &error);
         if (user_verifier == NULL) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -160,21 +284,46 @@ on_user_verifier_proxy_created (GObject            *source,
                                                    user_verifier,
                                                    (GDestroyNotify)
                                                    g_object_unref);
-        g_simple_async_result_complete_in_idle (operation_result);
+
+        self = GDM_CLIENT (g_async_result_get_source_object (G_ASYNC_RESULT (operation_result)));
+        if (self->priv->enabled_extensions == NULL) {
+                g_debug ("no enabled extensions");
+                g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
+                return;
+        }
+
+        self->priv->user_verifier_extensions = g_hash_table_new_full (g_str_hash,
+                                                                      g_str_equal,
+                                                                      NULL,
+                                                                      (GDestroyNotify)
+                                                                      free_interface_skeleton);
+        cancellable = g_object_get_data (G_OBJECT (operation_result), "cancellable");
+        gdm_user_verifier_call_enable_extensions (user_verifier,
+                                                  (const char * const *)
+                                                  self->priv->enabled_extensions,
+                                                  cancellable,
+                                                  (GAsyncReadyCallback)
+                                                  on_user_verifier_extensions_enabled,
+                                                  operation_result);
+
 }
 
 static void
-on_reauthentication_channel_connected (GDBusConnection    *connection,
+on_reauthentication_channel_connected (GObject            *source_object,
                                        GAsyncResult       *result,
                                        GSimpleAsyncResult *operation_result)
 {
+        GDBusConnection *connection;
         GCancellable *cancellable;
         GError       *error;
 
         error = NULL;
-        if (!g_dbus_connection_new_for_address_finish (result, &error)) {
+        connection = g_dbus_connection_new_for_address_finish (result, &error);
+        if (!connection) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -187,6 +336,7 @@ on_reauthentication_channel_connected (GDBusConnection    *connection,
                                      (GAsyncReadyCallback)
                                      on_user_verifier_proxy_created,
                                      operation_result);
+        g_object_unref (connection);
 }
 
 static void
@@ -205,6 +355,7 @@ on_reauthentication_channel_opened (GdmManager         *manager,
                                                                     &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -232,6 +383,7 @@ on_got_manager_for_reauthentication (GdmClient           *client,
                                                    &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -303,17 +455,19 @@ gdm_client_open_connection_sync (GdmClient      *client,
 }
 
 static void
-on_connected (GDBusConnection    *connection,
+on_connected (GObject            *source_object,
               GAsyncResult       *result,
               GSimpleAsyncResult *operation_result)
 {
+        GDBusConnection *connection;
         GError *error;
 
         error = NULL;
-        if (!g_dbus_connection_new_for_address_finish (result,
-                                                       &error)) {
+        connection = g_dbus_connection_new_for_address_finish (result, &error);
+        if (!connection) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -322,6 +476,8 @@ on_connected (GDBusConnection    *connection,
                                                    (GDestroyNotify)
                                                    g_object_unref);
         g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
+        g_object_unref (connection);
 }
 
 static void
@@ -342,6 +498,8 @@ on_session_opened (GdmManager         *manager,
                                                    &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
+                g_object_unref (client);
                 return;
         }
 
@@ -353,6 +511,7 @@ on_session_opened (GdmManager         *manager,
                                            (GAsyncReadyCallback)
                                            on_connected,
                                            operation_result);
+        g_object_unref (client);
 }
 
 static void
@@ -368,6 +527,7 @@ on_got_manager_for_opening_connection (GdmClient           *client,
                                                    &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -446,6 +606,7 @@ gdm_client_open_connection (GdmClient           *client,
                                                        (GDestroyNotify)
                                                        g_object_unref);
             g_simple_async_result_complete_in_idle (operation_result);
+            g_object_unref (operation_result);
             return;
         }
 
@@ -668,6 +829,39 @@ gdm_client_get_user_verifier_sync (GdmClient     *client,
                                    (GWeakNotify)
                                    g_clear_object,
                                    &client->priv->connection);
+
+                if (client->priv->enabled_extensions != NULL) {
+                        gboolean res;
+
+                        client->priv->user_verifier_extensions = g_hash_table_new_full (g_str_hash,
+                                                                                        g_str_equal,
+                                                                                        NULL,
+                                                                                        (GDestroyNotify)
+                                                                                        free_interface_skeleton);
+                        res = gdm_user_verifier_call_enable_extensions_sync (client->priv->user_verifier,
+                                                                            (const char * const *)
+                                                                             client->priv->enabled_extensions,
+                                                                             cancellable,
+                                                                             NULL);
+
+                        if (res) {
+                                size_t i;
+                                for (i = 0; client->priv->enabled_extensions[i] != NULL; i++) {
+                                            if (strcmp (client->priv->enabled_extensions[i],
+                                                        gdm_user_verifier_choice_list_interface_info ()->name) == 0) {
+                                                        GdmUserVerifierChoiceList *choice_list_interface;
+                                                        choice_list_interface = gdm_user_verifier_choice_list_proxy_new_sync (client->priv->connection,
+                                                                                                                              G_DBUS_PROXY_FLAGS_NONE,
+                                                                                                                              NULL,
+                                                                                                                              SESSION_DBUS_PATH,
+                                                                                                                              cancellable,
+                                                                                                                              NULL);
+                                                        if (choice_list_interface != NULL)
+                                                                    g_hash_table_insert (client->priv->user_verifier_extensions, client->priv->enabled_extensions[i], choice_list_interface);
+                                            }
+                                }
+                        }
+                }
         }
 
         return client->priv->user_verifier;
@@ -685,6 +879,7 @@ on_connection_opened_for_user_verifier (GdmClient          *client,
         if (!gdm_client_open_connection_finish (client, result, &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -735,6 +930,7 @@ gdm_client_get_user_verifier (GdmClient           *client,
                                                            (GDestroyNotify)
                                                            g_object_unref);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -793,6 +989,26 @@ gdm_client_get_user_verifier_finish (GdmClient       *client,
         return g_object_ref (user_verifier);
 }
 
+/**
+ * gdm_client_get_user_verifier_choice_list:
+ * @client: a #GdmClient
+ *
+ * Gets a #GdmUserVerifierChoiceList object that can be used to
+ * verify a user's local account.
+ *
+ * Returns: (transfer none): #GdmUserVerifierChoiceList or %NULL if user
+ * verifier isn't yet fetched, or daemon doesn't support choice lists
+ */
+GdmUserVerifierChoiceList *
+gdm_client_get_user_verifier_choice_list (GdmClient *client)
+{
+        if (client->priv->user_verifier_extensions == NULL)
+                return NULL;
+
+        return g_hash_table_lookup (client->priv->user_verifier_extensions,
+                                    gdm_user_verifier_choice_list_interface_info ()->name);
+}
+
 static void
 on_timed_login_details_got (GdmGreeter   *greeter,
                             GAsyncResult *result)
@@ -825,6 +1041,7 @@ on_greeter_proxy_created (GObject            *source,
         if (greeter == NULL) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -833,6 +1050,7 @@ on_greeter_proxy_created (GObject            *source,
                                                    (GDestroyNotify)
                                                    g_object_unref);
         g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
 
         query_for_timed_login_requested_signal (greeter);
 }
@@ -849,6 +1067,7 @@ on_connection_opened_for_greeter (GdmClient          *client,
         if (!gdm_client_open_connection_finish (client, result, &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -899,6 +1118,7 @@ gdm_client_get_greeter (GdmClient           *client,
                                                            (GDestroyNotify)
                                                            g_object_unref);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1021,6 +1241,7 @@ on_remote_greeter_proxy_created (GObject            *object,
         if (remote_greeter == NULL) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1029,6 +1250,7 @@ on_remote_greeter_proxy_created (GObject            *object,
                                                    (GDestroyNotify)
                                                    g_object_unref);
         g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
 }
 
 static void
@@ -1043,6 +1265,7 @@ on_connection_opened_for_remote_greeter (GdmClient          *client,
         if (!gdm_client_open_connection_finish (client, result, &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1093,6 +1316,7 @@ gdm_client_get_remote_greeter (GdmClient           *client,
                                                            (GDestroyNotify)
                                                            g_object_unref);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1212,6 +1436,7 @@ on_chooser_proxy_created (GObject            *source,
         if (chooser == NULL) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1220,6 +1445,7 @@ on_chooser_proxy_created (GObject            *source,
                                                    (GDestroyNotify)
                                                    g_object_unref);
         g_simple_async_result_complete_in_idle (operation_result);
+        g_object_unref (operation_result);
 }
 
 static void
@@ -1234,6 +1460,7 @@ on_connection_opened_for_chooser (GdmClient          *client,
         if (!gdm_client_open_connection_finish (client, result, &error)) {
                 g_simple_async_result_take_error (operation_result, error);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1284,6 +1511,7 @@ gdm_client_get_chooser (GdmClient           *client,
                                                            (GDestroyNotify)
                                                            g_object_unref);
                 g_simple_async_result_complete_in_idle (operation_result);
+                g_object_unref (operation_result);
                 return;
         }
 
@@ -1449,6 +1677,7 @@ gdm_client_finalize (GObject *object)
         g_clear_object (&client->priv->manager);
         g_clear_object (&client->priv->connection);
 
+        g_strfreev (client->priv->enabled_extensions);
         g_free (client->priv->address);
 
         G_OBJECT_CLASS (gdm_client_parent_class)->finalize (object);
@@ -1467,3 +1696,20 @@ gdm_client_new (void)
 
         return GDM_CLIENT (client_object);
 }
+
+
+/**
+ * gdm_client_set_enabled_extensions:
+ * @client: a #GdmClient
+ * @extensions: (array zero-terminated=1) (element-type utf8): a list of extensions
+ *
+ * Enables GDM's pam extensions.  Currently, only
+ * org.gnome.DisplayManager.UserVerifier.ChoiceList is supported.
+ */
+void
+gdm_client_set_enabled_extensions (GdmClient          *client,
+                                   const char * const *extensions)
+{
+        client->priv->enabled_extensions = g_strdupv ((char **) extensions);
+
+}
